
import numpy as np
import torch
from torch import optim
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

class Generator(nn.Module):
    def __init__(self, input_size, hidden_size, prev_timesteps, device, tonal_range, time_steps, dropout_prob=0.5):
        super(Generator, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        
        # Generate based on melody and previous output
        if prev_timesteps:
            self.linear1 = nn.Linear((2*input_size), hidden_size)

        # Generate based on melody alone
        else:
            self.linear1 = nn.Linear(input_size, hidden_size)
        
        self.lstm = nn.LSTM(hidden_size, hidden_size, num_layers=2, batch_first=True, dropout=dropout_prob)
        self.linear2 = nn.Linear(hidden_size, tonal_range)
        self.prev_timesteps = prev_timesteps 
        self.device = device
        self.tonal_range = tonal_range # 84 or 24
        self.time_steps = time_steps

    def forward(self, input, state, teacher_forcing_ratio = 0.5):
        input = input.reshape(1, 384, self.tonal_range)
        
        # Initialize empty previous time steps
        if self.time_steps > 1:
            prev_input = torch.empty(1, self.time_steps-1, self.tonal_range).uniform_().float().to(self.device)

            input = torch.cat((prev_input, input), dim=1)
       
        hidden, cell = state 
        gen_feats = []

        
        # Generate based on melody and previous output
        if self.prev_timesteps:
            prev_gen = torch.empty(1, self.time_steps, self.tonal_range).uniform_().float().to(self.device)

            for i in range(len(input[0])-self.time_steps+1):
                
                curr_z = input[:,i:i+self.time_steps]
                curr_gen = prev_gen
                concat_in = torch.cat((curr_gen, curr_z), dim=2)
                output = self.linear1(concat_in.reshape(1, 1, self.time_steps*self.tonal_range*2))
                output = F.relu(output)
                output, (hidden, cell) = self.lstm(output, (hidden, cell))
                output = self.linear2(output)
                prev_gen = torch.cat((prev_gen[:,1:self.time_steps+1], output.reshape(1, 1, self.tonal_range)), dim=1).reshape(1, self.time_steps,self.tonal_range)

                gen_feats.append(output)

        # Generate based on melody alone
        else:
            for i in range(len(input[0])-self.time_steps+1):
            
                curr_z = input[:,i:i+self.time_steps]
                output = self.linear1(curr_z.reshape(1, 1, self.time_steps*self.tonal_range))
                output = F.relu(output)
                output, (hidden, cell) = self.lstm(output, (hidden, cell))
                output = self.linear2(output)
                gen_feats.append(output)
        gen_feats = torch.stack(gen_feats).reshape(1, 384, self.tonal_range)
        
        state = (hidden, cell)
        return gen_feats, state
    
    def init_hidden(self):
        state = (Variable(torch.zeros(2, 1, self.hidden_size)).to(self.device),
                Variable(torch.zeros(2, 1, self.hidden_size)).to(self.device))
        return state



class Discriminator(nn.Module):
    def __init__(self, hidden_size, output_size, device, time_steps, tonal_range, dropout_prob=0.5):
        super(Discriminator, self).__init__()
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.linear1 = nn.Linear(in_features=output_size, out_features=hidden_size)
        self.lstm = nn.LSTM(hidden_size, hidden_size, num_layers = 2, batch_first=True, bidirectional=True, dropout=dropout_prob)
        self.linear2 = nn.Linear(in_features= 2 * hidden_size, out_features=1)
        self.time_steps = time_steps
        self.device = device
        self.tonal_range = tonal_range

    def forward(self, input, state):
        hidden, cell = state
        input = input.reshape(1, 384, self.tonal_range)
        if self.time_steps > 1:
            prev_input = torch.empty([1, self.time_steps-1, self.tonal_range]).uniform_().float().to(self.device)
            input = torch.cat((prev_input, input), dim=1)
        outputs = []
        for i in range(len(input[0])-self.time_steps+1):
            curr_z = input[:,i:i+self.time_steps]
            lin_out = self.linear1(curr_z.reshape(1, 1, self.time_steps*self.tonal_range))
            lstm_output, state = self.lstm(lin_out, (hidden, cell))
            output = self.linear2(lstm_output)
            outputs.append(output)
            hidden, cell = state
        outputs = torch.stack(outputs)
        state = (hidden, cell)
        return outputs, lstm_output, state

    def init_hidden(self):
        state = (Variable(torch.zeros(4, 1, self.hidden_size)).to(self.device),
                Variable(torch.zeros(4, 1, self.hidden_size)).to(self.device))
        return state
