import numpy as np
import torch
import random
import pypianoroll
import os
from matplotlib import pyplot as plt


def plot_results(g_loss_arr, d_loss_arr, total_corr_arr, total_acc_arr, pretrain, config):
    fig = plt.figure(figsize=(10, 7), dpi=80)
    pretrain_coords = []

    # Add pretraining statistics
    if config["pretraining"]:
        pretrain_coords.append(len(pretrain['g_loss_fg']))
        pretrain_coords.append((len(pretrain['g_loss_fg']) * 2))
        
        pretrain_g = np.append(pretrain['g_loss_fg'], pretrain['g_loss_fd'])
        pretrain_d = np.append(pretrain['d_loss_fg'], pretrain['d_loss_fd'])
        
        pretrain_d_acc = np.append(pretrain['d_acc_fg'], pretrain['d_acc_fd'])
        pretrain_d_corr = np.append(pretrain['d_corr_fg'], pretrain['d_corr_fd'])
        
        g_loss_arr = np.append(pretrain_g, g_loss_arr)
        d_loss_arr = np.append(pretrain_d, d_loss_arr)
        total_acc_arr = np.append(pretrain_d_acc, total_acc_arr)
        total_corr_arr = np.append(pretrain_d_corr, total_corr_arr)
    
    # Add lines marking the pretraining epochs
    for xc in pretrain_coords:
        plt.axvline(x=xc, color='r')

    plt.plot(g_loss_arr, label="Generator", linewidth=2.0)
    plt.plot(d_loss_arr, label="Discriminator", linewidth=2.0)
    plt.xlabel("Epochs")
    plt.legend(loc="best")
    
    plt.savefig('../output/losses.png')

    fig = plt.figure(figsize=(10,7), dpi = 80)
    plt.xlabel("Epochs")
    plt.ylabel("Accuracy")

    # Add lines marking the pretraining epochs
    for xc in pretrain_coords:
        plt.axvline(x=xc, color='r')
    plt.plot(total_acc_arr, linewidth=2.0)
    plt.savefig('../output/accuracy.png')

    fig = plt.figure(figsize=(10,7), dpi = 80)
    plt.xlabel("Epochs")
    plt.ylabel("Correctly labeled")

    # Add lines marking the pretraining epochs
    for xc in pretrain_coords:
        plt.axvline(x=xc, color='r')
    plt.grid(color='#95a5a6', linestyle='--', linewidth=1, axis='y', alpha=0.6)
    plt.bar(range(len(total_corr_arr)), height=total_corr_arr, align="center")
    plt.savefig("../output/numcorrects.png")
    plt.show()
    
def save_samples_midi(generator, data, config):
    ih = generator.init_hidden()

    # Indices of the random samples in the test data
    sample_idx = random.sample(range(0, len(data["test_x"])-1), config["gen_samples"])

    # Create output directory
    os.makedirs("../output")
    for i in range(0, len(sample_idx)):
        
        # Generate accompaniment for sample from test set
        input_phrase = data['test_x'][sample_idx[i]]
        gen,_ = generator(input_phrase, ih)
        temp = np.zeros((384, 22))
        
        # Transform into notes and chords
        chord_arr = get_generated_chords(gen)
        
        # Transform all time steps into the range 128 used by pypianoroll
        chord_arr = np.append(temp, chord_arr, axis=1)
        chord_arr = np.append(chord_arr, temp, axis=1)
        
        # Get the leading melody and transform all time steps into range 128
        melody = data["test_x_org"][sample_idx[i]]
        melody = melody.reshape(384, 84)
        melody = np.append(temp, melody, axis=1)
        melody = np.append(melody, temp, axis=1)
        melody = melody * 80.

        # Add the generated accompaniment and melody to a multitrack and write to a .mid-file
        melody = pypianoroll.Track(melody, name="Training data")
        acc = pypianoroll.Track(chord_arr, name="Generated")          
        multitrack = pypianoroll.Multitrack(tracks=[acc])
        multitrack.write("../output/gen_acc{}_isolated.mid".format(i+1))
        
        multitrack = pypianoroll.Multitrack(tracks=[melody, acc])
        multitrack.write("../output/gen_combined{}.mid".format(i+1))
        
        multitrack = pypianoroll.Multitrack(tracks=[melody])
        multitrack.write("../output/melody{}_isolated.mid".format(i+1))
        
        # Get the real accompaniment and create a .mid file with it and the melody
        track = data['test_y_org'][sample_idx[i]]
        track = track.reshape(384, 84)
        track = np.append(temp, track, axis=1)
        track = np.append(track, temp, axis=1)
        track = track * 100.
        track = pypianoroll.Track(track, name="Real")
        
        multitrack = pypianoroll.Multitrack(tracks=[track])
        multitrack.write("../output/real_acc{}_isolated.mid".format(i+1))

        multitrack = pypianoroll.Multitrack(tracks=[melody, track])
        multitrack.write("../output/real_combined{}.mid".format(i+1))


def write_to_file(data, name, config):

    red = "r" if config["reduce_dim"] else "f"
    ts = config["time_steps"]
    prev = "p" if config["prev_timesteps"] else "l"
    f = open("../output/{}.txt".format(name), "w")
    for i in data:
        f.write("\n")
        f.write(str(i))
    f.close()

def get_generated_chords(gen_bars):
    
    gen_bars = torch.squeeze(gen_bars).detach().cpu().numpy()
    
    chords = np.zeros((len(gen_bars), 84))
    
    for i, bar in enumerate(gen_bars):
        # Gets the note with the highest probability
        max_val = np.argmax(bar)

        # Most probable note gets 100 % in velocity
        chords[i][max_val] = 100.
        
    return chords

def reduce_data_dim(data):
    training_data = np.zeros((len(data), 384, 24))
    
    # Reduce tonal range from 84 to 24
    for i in range(len(data)):
        for j in range(len(data[i])):
            max_idx = np.argwhere(data[i][j] == 1.)
            max_idx = max_idx % 24
            training_data[i][j][max_idx] = 1.
    return training_data

def load_data(filepath, num_samples, device, test_size=0.2, reduce_dim = False):
    print("Loading data...")
    data = np.load(filepath)

    # Random sampling
    rand_idx = random.sample(range(0, len(data)-1), num_samples)

    sampled_data = data[rand_idx]
    
    mask = np.ones(len(sampled_data))
    
    # Random sampling for test data
    test_idx = random.sample(range(0, len(sampled_data)-1), int(np.round(num_samples * test_size)))
    
    mask[test_idx] = 0.
    
    keep_idx = np.where(mask == 1.)
    
    train = sampled_data[keep_idx]
    test = sampled_data[test_idx]

    test_x = test[:,:,:,:,2]
    test_y = test[:,:,:,:,1]

    # Reshape to proper piano-roll representation
    test_x_org = test_x.reshape(len(test_x), 384, 84)
    test_y_org = test_y.reshape(len(test_y), 384, 84)

    # Divide the data into melody and accompaniment
    training_data = train[:, :, :, :,2]
    labels = train[:, :, :, :,1]

    training_data = training_data.reshape(len(training_data), 384, 84)
    labels = labels.reshape(len(labels), 384, 84)
    
    # Reduce from 84 to 24
    if reduce_dim:
        training_data = reduce_data_dim(training_data)
        labels = reduce_data_dim(labels)

        test_x = reduce_data_dim(test_x_org)
        test_y = reduce_data_dim(test_y_org)
    else:
        test_x = test_x_org
        test_y = test_y_org

    train_x = torch.from_numpy(training_data).float().to(device)
    train_y = torch.from_numpy(labels).float().to(device)

    test_x = torch.from_numpy(test_x).float().to(device)
    test_y = torch.from_numpy(test_y).float().to(device)
    
    data_dict = {
        "train_x": train_x,
        "train_y": train_y,
        "test_x": test_x,
        "test_y": test_y,
        "test_x_org": test_x_org,
        "test_y_org": test_y_org
    }
    
    print("Data loaded!")
    return data_dict